@@ -31,11 +31,12 @@
  */
 
 #include "ucm_local.h"
+#include <sys/stat.h>
 #include <stdbool.h>
 #include <dirent.h>
 #include <limits.h>
 
-static int filename_filter(const struct dirent *dirent);
+static int filename_filter(const struct dirent64 *dirent);
 
 static int parse_sequence(snd_use_case_mgr_t *uc_mgr,
 			  struct list_head *base,
@@ -239,7 +240,8 @@ static int error_node(snd_use_case_mgr_t *uc_mgr, snd_config_t *cfg)
 		uc_error("error: failed to get Error string");
 		return err;
 	}
-	uc_error("%s", s);
+	if (!uc_mgr->suppress_nodev_errors)
+		uc_error("%s", s);
 	free(s);
 	return -ENXIO;
 }
@@ -276,6 +278,10 @@ static int evaluate_regex(snd_use_case_mgr_t *uc_mgr,
 		err = snd_config_get_id(n, &id);
 		if (err < 0)
 			return err;
+		if (id[0] == '@') {
+			uc_error("error: value names starting with '@' are reserved for application variables");
+			return -EINVAL;
+		}
 		err = uc_mgr_define_regex(uc_mgr, id, n);
 		if (err < 0)
 			return err;
@@ -325,8 +331,15 @@ static int evaluate_define(snd_use_case_mgr_t *uc_mgr,
 		free(var);
 		if (err < 0)
 			return err;
-		uc_mgr_set_variable(uc_mgr, id, s);
+		if (id[0] == '@') {
+			free(s);
+			uc_error("error: value names starting with '@' are reserved for application variables");
+			return -EINVAL;
+		}
+		err = uc_mgr_set_variable(uc_mgr, id, s);
 		free(s);
+		if (err < 0)
+			return err;
 	}
 
 	snd_config_delete(d);
@@ -334,6 +347,167 @@ static int evaluate_define(snd_use_case_mgr_t *uc_mgr,
 	return evaluate_regex(uc_mgr, cfg);
 }
 
+/*
+ * Evaluate macro definitions (in-place delete)
+ */
+static int evaluate_define_macro(snd_use_case_mgr_t *uc_mgr,
+				 snd_config_t *cfg)
+{
+	snd_config_t *d;
+	int err;
+
+	err = snd_config_search(cfg, "DefineMacro", &d);
+	if (err == -ENOENT)
+		return 1;
+	if (err < 0)
+		return err;
+
+	if (snd_config_get_type(d) != SND_CONFIG_TYPE_COMPOUND) {
+		uc_error("compound type expected for DefineMacro");
+		return -EINVAL;
+	}
+
+	if (uc_mgr->conf_format < 6) {
+		uc_error("DefineMacro is supported in v6+ syntax");
+		return -EINVAL;
+	}
+
+	err = snd_config_merge(uc_mgr->macros, d, 0);
+	if (err < 0)
+		return err;
+	return 0;
+}
+
+static int evaluate_macro1(snd_use_case_mgr_t *uc_mgr,
+			   snd_config_t *dst,
+			   snd_config_t *args)
+{
+	snd_config_iterator_t i, next;
+	snd_config_t *m, *mc, *a, *n;
+	const char *mid, *id;
+	char name[128], *var;
+	const char *s;
+	int err;
+
+	err = snd_config_get_id(args, &mid);
+	if (err < 0)
+		return err;
+	err = snd_config_search(uc_mgr->macros, mid, &m);
+	if (err < 0) {
+		uc_error("Macro '%s' is not defined", mid);
+		return err;
+	}
+
+	a = args;
+	if (snd_config_get_type(args) == SND_CONFIG_TYPE_STRING) {
+		err = snd_config_get_string(args, &s);
+		if (err < 0)
+			return err;
+		err = snd_config_load_string(&a, s, 0);
+		if (err < 0)
+			return err;
+	} else if (snd_config_get_type(args) != SND_CONFIG_TYPE_COMPOUND) {
+		return -EINVAL;
+	}
+
+	/* set arguments */
+	snd_config_for_each(i, next, a) {
+		n = snd_config_iterator_entry(i);
+		err = snd_config_get_id(n, &id);
+		if (err < 0)
+			return err;
+		err = snd_config_get_ascii(n, &var);
+		if (err < 0)
+			return err;
+		snprintf(name, sizeof(name), "__%s", id);
+		err = uc_mgr_set_variable(uc_mgr, name, var);
+		free(var);
+		if (err < 0)
+			return err;
+	}
+
+	/* merge + substitute variables */
+	err = snd_config_copy(&mc, m);
+	if (err < 0)
+		goto __err_path;
+	err = uc_mgr_evaluate_inplace(uc_mgr, mc);
+	if (err < 0) {
+		snd_config_delete(mc);
+		goto __err_path;
+	}
+	err = uc_mgr_config_tree_merge(uc_mgr, dst, mc, NULL, NULL);
+	snd_config_delete(mc);
+
+	/* delete arguments */
+	snd_config_for_each(i, next, a) {
+		n = snd_config_iterator_entry(i);
+		err = snd_config_get_id(n, &id);
+		if (err < 0)
+			return err;
+		snprintf(name, sizeof(name), "__%s", id);
+		err = uc_mgr_delete_variable(uc_mgr, name);
+		if (err < 0)
+			return err;
+	}
+
+__err_path:
+	if (a != args)
+		snd_config_delete(a);
+	return err;
+}
+
+/*
+ * Evaluate macro definitions and instances (in-place delete)
+ */
+static int evaluate_macro(snd_use_case_mgr_t *uc_mgr,
+			  snd_config_t *cfg)
+{
+	snd_config_iterator_t i, i2, next, next2;
+	snd_config_t *d, *n, *n2;
+	int err, ret;
+
+	ret = evaluate_define_macro(uc_mgr, cfg);
+	if (ret < 0)
+		return ret;
+
+	err = snd_config_search(cfg, "Macro", &d);
+	if (err == -ENOENT)
+		return ret;
+	if (err < 0)
+		return err;
+
+	if (snd_config_get_type(d) != SND_CONFIG_TYPE_COMPOUND) {
+		uc_error("compound type expected for DefineMacro");
+		return -EINVAL;
+	}
+
+	if (uc_mgr->conf_format < 6) {
+		uc_error("Macro is supported in v6+ syntax");
+		return -EINVAL;
+	}
+
+	snd_config_for_each(i, next, d) {
+		n = snd_config_iterator_entry(i);
+		if (snd_config_get_type(n) != SND_CONFIG_TYPE_COMPOUND) {
+			const char *id;
+			if (snd_config_get_id(n, &id))
+				id = "";
+			uc_error("compound type expected for Macro.%s", id);
+			return -EINVAL;
+		}
+		snd_config_for_each(i2, next2, n) {
+			n2 = snd_config_iterator_entry(i2);
+			err = evaluate_macro1(uc_mgr, cfg, n2);
+			if (err < 0)
+				return err;
+		}
+	}
+
+	snd_config_delete(d);
+
+	return 0;
+}
+
 /*
  * Evaluate include (in-place)
  */
@@ -357,8 +531,7 @@ static int evaluate_include(snd_use_case_mgr_t *uc_mgr,
 /*
  * Evaluate condition (in-place)
  */
-static int evaluate_condition(snd_use_case_mgr_t *uc_mgr,
-			      snd_config_t *cfg)
+static int evaluate_condition(snd_use_case_mgr_t *uc_mgr, snd_config_t *cfg)
 {
 	snd_config_t *n;
 	int err;
@@ -374,15 +547,73 @@ static int evaluate_condition(snd_use_case_mgr_t *uc_mgr,
 	return err;
 }
 
+/*
+ * Evaluate variant (in-place)
+ */
+static int evaluate_variant(snd_use_case_mgr_t *uc_mgr, snd_config_t *cfg)
+{
+	snd_config_iterator_t i, next;
+	snd_config_t *n, *c;
+	const char *id;
+	int err;
+
+	err = snd_config_search(cfg, "Variant", &c);
+	if (err == -ENOENT)
+		return 1;
+	if (err < 0)
+		return err;
+
+	if (uc_mgr->conf_format < 6) {
+		uc_error("Variant is supported in v6+ syntax");
+		return -EINVAL;
+	}
+
+	if (uc_mgr->parse_master_section)
+		return 1;
+
+	if (uc_mgr->parse_variant == NULL)
+		goto __ret;
+
+	snd_config_for_each(i, next, c) {
+		n = snd_config_iterator_entry(i);
+
+		if (snd_config_get_id(n, &id) < 0)
+			return -EINVAL;
+
+		if (strcmp(id, uc_mgr->parse_variant))
+			continue;
+
+		err = uc_mgr_evaluate_inplace(uc_mgr, n);
+		if (err < 0)
+			return err;
+
+		err = uc_mgr_config_tree_merge(uc_mgr, cfg, n, NULL, NULL);
+		if (err < 0)
+			return err;
+		snd_config_delete(c);
+		return 0;
+	}
+
+__ret:
+	snd_config_delete(c);
+	return 1;
+}
+
 /*
  * In-place evaluate
  */
 int uc_mgr_evaluate_inplace(snd_use_case_mgr_t *uc_mgr,
 			    snd_config_t *cfg)
 {
-	int err1 = 0, err2 = 0, err3 = 0;
+	long iterations = 10000;
+	int err1 = 0, err2 = 0, err3 = 0, err4 = 0, err5 = 0;
 
-	while (err1 == 0 || err2 == 0 || err3 == 0) {
+	while (err1 == 0 || err2 == 0 || err3 == 0 || err4 == 0 || err5 == 0) {
+		if (iterations == 0) {
+			uc_error("Maximal inplace evaluation iterations number reached (recursive references?)");
+			return -EINVAL;
+		}
+		iterations--;
 		/* variables at first */
 		err1 = evaluate_define(uc_mgr, cfg);
 		if (err1 < 0)
@@ -391,13 +622,29 @@ int uc_mgr_evaluate_inplace(snd_use_case_mgr_t *uc_mgr,
 		err2 = evaluate_include(uc_mgr, cfg);
 		if (err2 < 0)
 			return err2;
-		/* include may define another variables */
+		/* include or macro may define another variables */
 		/* conditions may depend on them */
 		if (err2 == 0)
 			continue;
-		err3 = evaluate_condition(uc_mgr, cfg);
+		err3 = evaluate_variant(uc_mgr, cfg);
 		if (err3 < 0)
 			return err3;
+		if (err3 == 0)
+			continue;
+		uc_mgr->macro_hops++;
+		if (uc_mgr->macro_hops > 100) {
+			uc_error("Maximal macro hops reached!");
+			return -EINVAL;
+		}
+		err4 = evaluate_macro(uc_mgr, cfg);
+		uc_mgr->macro_hops--;
+		if (err4 < 0)
+			return err4;
+		if (err4 == 0)
+			continue;
+		err5 = evaluate_condition(uc_mgr, cfg);
+		if (err5 < 0)
+			return err5;
 	}
 	return 0;
 }
@@ -829,30 +1076,41 @@ cset:
 			continue;
 		}
 
-		if (strcmp(cmd, "enadev") == 0) {
-			/* need to enable a component device */
+		if (strcmp(cmd, "enadev") == 0 ||
+		    strcmp(cmd, "disdev") == 0) {
+			/* need to enable or disable a component device */
 			curr->type = SEQUENCE_ELEMENT_TYPE_CMPT_SEQ;
-			err = parse_component_seq(uc_mgr, n, 1,
+			err = parse_component_seq(uc_mgr, n,
+						strcmp(cmd, "enadev") == 0,
 						&curr->data.cmpt_seq);
 			if (err < 0) {
-				uc_error("error: enadev requires a valid device!");
+				uc_error("error: %s requires a valid device!", cmd);
 				return err;
 			}
 			continue;
 		}
 
-		if (strcmp(cmd, "disdev") == 0) {
-			/* need to disable a component device */
-			curr->type = SEQUENCE_ELEMENT_TYPE_CMPT_SEQ;
-			err = parse_component_seq(uc_mgr, n, 0,
-						&curr->data.cmpt_seq);
+		if (strcmp(cmd, "enadev2") == 0) {
+			curr->type = SEQUENCE_ELEMENT_TYPE_DEV_ENABLE_SEQ;
+			goto device;
+		}
+
+		if (strcmp(cmd, "disdev2") == 0) {
+			curr->type = SEQUENCE_ELEMENT_TYPE_DEV_DISABLE_SEQ;
+device:
+			err = parse_string_substitute3(uc_mgr, n, &curr->data.device);
 			if (err < 0) {
-				uc_error("error: disdev requires a valid device!");
+				uc_error("error: %s requires a valid device!", cmd);
 				return err;
 			}
 			continue;
 		}
 
+		if (strcmp(cmd, "disdevall") == 0) {
+			curr->type = SEQUENCE_ELEMENT_TYPE_DEV_DISABLE_ALL;
+			continue;
+		}
+
 		if (strcmp(cmd, "cset-bin-file") == 0) {
 			curr->type = SEQUENCE_ELEMENT_TYPE_CSET_BIN_FILE;
 			goto cset;
@@ -1469,7 +1727,7 @@ static int parse_modifier_name(snd_use_case_mgr_t *uc_mgr,
 			     void *data1,
 			     void *data2 ATTRIBUTE_UNUSED)
 {
-	return parse_compound(uc_mgr, cfg, parse_modifier, data1, data2);
+	return parse_compound_check_legacy(uc_mgr, cfg, parse_modifier, data1);
 }
 
 static int verb_dev_list_add(struct use_case_verb *verb,
@@ -1807,6 +2065,65 @@ static int parse_verb_file(snd_use_case_mgr_t *uc_mgr,
 	return err;
 }
 
+/*
+ * Parse variant information
+ */
+static int parse_variant(snd_use_case_mgr_t *uc_mgr, snd_config_t *cfg,
+			 char **_vfile, char **_vcomment)
+{
+	snd_config_iterator_t i, next;
+	snd_config_t *n;
+	char *file = NULL, *comment = NULL;
+	int err;
+
+	/* parse master config sections */
+	snd_config_for_each(i, next, cfg) {
+		const char *id;
+		n = snd_config_iterator_entry(i);
+		if (snd_config_get_id(n, &id) < 0)
+			continue;
+
+		/* get use case verb file name */
+		if (strcmp(id, "File") == 0) {
+			if (_vfile) {
+				err = parse_string_substitute3(uc_mgr, n, &file);
+				if (err < 0) {
+					uc_error("failed to get File");
+					goto __error;
+				}
+			}
+			continue;
+		}
+
+		/* get optional use case comment */
+		if (strncmp(id, "Comment", 7) == 0) {
+			if (_vcomment) {
+				err = parse_string_substitute3(uc_mgr, n, &comment);
+				if (err < 0) {
+					uc_error("error: failed to get Comment");
+					goto __error;
+				}
+			}
+			continue;
+		}
+
+		uc_error("unknown field '%s' in Variant section", id);
+		err = -EINVAL;
+		goto __error;
+	}
+
+	if (_vfile)
+		*_vfile = file;
+	if (_vcomment)
+		*_vcomment = comment;
+	return 0;
+
+__error:
+	free(file);
+	free(comment);
+	return err;
+}
+
 /*
  * Parse master section for "Use Case" and "File" tags.
  */
@@ -1815,8 +2132,9 @@ static int parse_master_section(snd_use_case_mgr_t *uc_mgr, snd_config_t *cfg,
 				void *data2 ATTRIBUTE_UNUSED)
 {
 	snd_config_iterator_t i, next;
-	snd_config_t *n;
+	snd_config_t *n, *variant = NULL;
 	char *use_case_name, *file = NULL, *comment = NULL;
+	bool variant_ok = false;
 	int err;
 
 	if (snd_config_get_type(cfg) != SND_CONFIG_TYPE_COMPOUND) {
@@ -1831,7 +2149,9 @@ static int parse_master_section(snd_use_case_mgr_t *uc_mgr, snd_config_t *cfg,
 	}
 
 	/* in-place evaluation */
+	uc_mgr->parse_master_section = 1;
 	err = uc_mgr_evaluate_inplace(uc_mgr, cfg);
+	uc_mgr->parse_master_section = 0;
 	if (err < 0)
 		goto __error;
 
@@ -1862,20 +2182,69 @@ static int parse_master_section(snd_use_case_mgr_t *uc_mgr, snd_config_t *cfg,
 			continue;
 		}
 
-		uc_error("unknown field %s in master section");
+		if (uc_mgr->conf_format >= 6 && strcmp(id, "Variant") == 0) {
+			snd_config_iterator_t i2, next2;
+			variant = n;
+			snd_config_for_each(i2, next2, n) {
+				const char *id2;
+				snd_config_t *n2;
+				n2 = snd_config_iterator_entry(i2);
+				if (snd_config_get_id(n2, &id2) < 0)
+					continue;
+				err = uc_mgr_evaluate_inplace(uc_mgr, n2);
+				if (err < 0)
+					goto __error;
+				if (strcmp(use_case_name, id2) == 0)
+					variant_ok = true;
+			}
+			continue;
+		}
+
+		uc_error("unknown field '%s' in SectionUseCase", id);
 	}
 
-	uc_dbg("use_case_name %s file '%s'", use_case_name, file);
-
-	/* do we have both use case name and file ? */
-	if (!file) {
-		uc_error("error: use case missing file");
+	if (variant && !variant_ok) {
+		uc_error("error: undefined variant '%s'", use_case_name);
 		err = -EINVAL;
 		goto __error;
 	}
 
-	/* parse verb file */
-	err = parse_verb_file(uc_mgr, use_case_name, comment, file);
+	if (!variant) {
+		uc_dbg("use_case_name %s file '%s'", use_case_name, file);
+
+		/* do we have both use case name and file ? */
+		if (!file) {
+			uc_error("error: use case missing file");
+			err = -EINVAL;
+			goto __error;
+		}
+
+		/* parse verb file */
+		err = parse_verb_file(uc_mgr, use_case_name, comment, file);
+	} else {
+		/* parse variants */
+		snd_config_for_each(i, next, variant) {
+			char *vfile, *vcomment;
+			const char *id;
+			n = snd_config_iterator_entry(i);
+			if (snd_config_get_id(n, &id) < 0)
+				continue;
+			if (!parse_is_name_safe(id)) {
+				err = -EINVAL;
+				goto __error;
+			}
+			err = parse_variant(uc_mgr, n, &vfile, &vcomment);
+			if (err < 0)
+				break;
+			uc_mgr->parse_variant = id;
+			err = parse_verb_file(uc_mgr, id,
+						vcomment ? vcomment : comment,
+						vfile ? vfile : file);
+			uc_mgr->parse_variant = NULL;
+			free(vfile);
+			free(vcomment);
+		}
+	}
 
 __error:
 	free(use_case_name);
@@ -2185,6 +2554,7 @@ static int parse_toplevel_path(snd_use_case_mgr_t *uc_mgr,
 	snd_config_t *n, *n2;
 	const char *id;
 	char *dir = NULL, *file = NULL, fn[PATH_MAX];
+	struct stat64 st;
 	long version;
 	int err;
 
@@ -2259,23 +2629,51 @@ static int parse_toplevel_path(snd_use_case_mgr_t *uc_mgr,
 		}
 
 		ucm_filename(fn, sizeof(fn), version, dir, file);
-		if (access(fn, R_OK) == 0) {
-			if (replace_string(&uc_mgr->conf_dir_name, dir) == NULL) {
-				err = -ENOMEM;
-				goto __error;
-			}
-			if (replace_string(&uc_mgr->conf_file_name, file) == NULL) {
-				err = -ENOMEM;
-				goto __error;
+		if (access(fn, R_OK) == 0 && lstat64(fn, &st) == 0) {
+			if (st.st_mode & S_IFLNK) {
+				ssize_t r;
+				char *link, *dir2, *p;
+
+				link = malloc(PATH_MAX);
+				if (link == NULL)
+					goto __enomem;
+				r = readlink(fn, link, PATH_MAX - 1);
+				if (r <= 0) {
+					free(link);
+					goto __next;
+				}
+				link[r] = '\0';
+				p = strrchr(link, '/');
+				if (p) {
+					*p = '\0';
+					dir2 = malloc(PATH_MAX);
+					if (dir2 == NULL) {
+						free(link);
+						goto __enomem;
+					}
+					strncpy(dir2, dir, PATH_MAX - 1);
+					strncat(dir2, "/", PATH_MAX - 1);
+					strncat(dir2, link, PATH_MAX - 1);
+					fn[PATH_MAX - 1] = '\0';
+					free(dir);
+					dir = dir2;
+				}
+				free(link);
 			}
+			if (replace_string(&uc_mgr->conf_dir_name, dir) == NULL)
+				goto __enomem;
+			if (replace_string(&uc_mgr->conf_file_name, file) == NULL)
+				goto __enomem;
 			strncpy(filename, fn, PATH_MAX);
+			filename[PATH_MAX - 1] = '\0';
 			uc_mgr->conf_format = version;
 			goto __ok;
 		}
 
 __next:
 		free(file);
-		free(dir);
+		if (dir != fn)
+			free(dir);
 		dir = NULL;
 		file = NULL;
 	}
@@ -2283,11 +2681,16 @@ __next:
 	err = -ENOENT;
 	goto __error;
 
+__enomem:
+	err = -ENOMEM;
+	goto __error;
+
 __ok:
 	err = 0;
 __error:
 	free(file);
-	free(dir);
+	if (dir != fn)
+		free(dir);
 	return err;
 }
 
@@ -2404,6 +2807,14 @@ int uc_mgr_import_master_config(snd_use_case_mgr_t *uc_mgr)
 	const char *name;
 	int err;
 
+	err = snd_config_top(&uc_mgr->local_config);
+	if (err < 0)
+		return err;
+
+	err = snd_config_top(&uc_mgr->macros);
+	if (err < 0)
+		return err;
+
 	name = uc_mgr->card_name;
 	if (strncmp(name, "hw:", 3) == 0) {
 		err = get_by_card(uc_mgr, name);
@@ -2422,6 +2833,10 @@ int uc_mgr_import_master_config(snd_use_case_mgr_t *uc_mgr)
 		goto __error;
 
 	err = parse_master_file(uc_mgr, cfg);
+	if (uc_mgr->macros) {
+		snd_config_delete(uc_mgr->macros);
+		uc_mgr->macros = NULL;
+	}
 	snd_config_delete(cfg);
 	if (err < 0) {
 		uc_mgr_free_ctl_list(uc_mgr);
@@ -2436,7 +2851,7 @@ __error:
 	return err;
 }
 
-static int filename_filter(const struct dirent *dirent)
+static int filename_filter(const struct dirent64 *dirent)
 {
 	if (dirent == NULL)
 		return 0;
@@ -2470,7 +2885,7 @@ int uc_mgr_scan_master_configs(const char **_list[])
 	int i, j, cnt, err;
 	long l;
 	ssize_t ss;
-	struct dirent **namelist;
+	struct dirent64 **namelist;
 
 	if (env)
 		snprintf(filename, sizeof(filename), "%s/conf.virt.d", env);
@@ -2479,11 +2894,11 @@ int uc_mgr_scan_master_configs(const char **_list[])
 			 snd_config_topdir());
 
 #if defined(_GNU_SOURCE) && !defined(__NetBSD__) && !defined(__FreeBSD__) && !defined(__sun) && !defined(ANDROID)
-#define SORTFUNC	versionsort
+#define SORTFUNC	versionsort64
 #else
-#define SORTFUNC	alphasort
+#define SORTFUNC	alphasort64
 #endif
-	err = scandir(filename, &namelist, filename_filter, SORTFUNC);
+	err = scandir64(filename, &namelist, filename_filter, SORTFUNC);
 	if (err < 0) {
 		err = -errno;
 		uc_error("error: could not scan directory %s: %s",
